/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include <gio/gio.h>
#include <gio/giotypes.h>
#include <stdio.h>

#include <condition_variable>
#include <filesystem>
#include <fstream>
#include <future>
#include <iostream>
#include <mutex>
#include <thread>
#include <vector>

#include "embeddingcommon.h"
#include "kylin-ai/coreai/embedding/embedding.h"
#include "kylin-ai/coreai/embedding/error.h"

void callback(EmbeddingResult *result, void *callback_user_data) {
    // 结果信息
    int error_code = embedding_result_get_error_code(result);
    std::cout << "error_code: " << error_code << std::endl;
    const char *error_message = embedding_result_get_error_message(result);
    std::cout << "error_message :" << error_message << std::endl;

    float *vector_result = embedding_result_get_vector_data(result);
    int len = embedding_result_get_vector_length(result);

    std::cout << "vector_result :" << std::endl;
    for (int i = 0; i < len; i++) {
        std::cout << vector_result[i] << " ";
    }
    std::cout << std::endl;

    if (callback_user_data != nullptr) {
        int *a = static_cast<int *>(callback_user_data);
        std::cout << "user_data :" << *a << std::endl;
    }

    std::cout << "-----------------------------------" << std::endl;
    std::cout << "释放结果资源" << std::endl;
}
gboolean quit_loop(gpointer data) {
    GMainLoop *loop = static_cast<GMainLoop *>(data);
    g_main_loop_quit(loop);
    return G_SOURCE_REMOVE;  // 返回 G_SOURCE_REMOVE 以移除定时器
}

void textEmbeddingSync() {
    TextEmbeddingSession *session = text_embedding_create_session();
    int initSession = text_embedding_init_session(session);
    if (initSession != 0) {
        std::cout << "init session failed" << std::endl;
    }
    std::cout << "initSession :" << initSession << std::endl;
    // 同步接口
    EmbeddingResult *result = nullptr;
    bool sucess = text_embedding(session, "12345", &result);

    // 结果信息
    int error_code = embedding_result_get_error_code(result);
    std::cout << "error_code: " << error_code << std::endl;
    const char *error_message = embedding_result_get_error_message(result);
    std::cout << "error_message :" << error_message << std::endl;

    float *vector_result = embedding_result_get_vector_data(result);
    int len = embedding_result_get_vector_length(result);

    std::cout << "vector_result :" << std::endl;
    for (int i = 0; i < len; i++) {
        std::cout << vector_result[i] << ",";
    }
    std::cout << std::endl;

    char *info = nullptr;
    text_embedding_get_model_info(session, &info);
    std::cout << "print model info:" << info << std::endl;

#if 0
    // 测试重连
    GMainLoop *loop = g_main_loop_new(nullptr, false);
    g_timeout_add_seconds(5, quit_loop, loop);
    g_main_loop_run(loop);

    EmbeddingResult *result1 = nullptr;
    text_embedding(session, "12345", &result1);

    int error_code1 = embedding_result_get_error_code(result1);
    std::cout << "error_code1: " << error_code1 << std::endl;
    const char *error_message1 = embedding_result_get_error_message(result1);
    std::cout << "error_message1 :" << error_message1 << std::endl;

    if (error_code1 == COREAI_EMBEDDING_CONNECTION_ERROR) {
        // 错误码为10 说明连接runtime有问题
        embedding_result_destroy(&result);
        text_embedding_destroy_session(&session);
        //  由于runtime的问题，所以 重新创建session,重连两次，每次等待10s
        TextEmbeddingSession *sessionRe = text_embedding_create_session();
        int initSessionRe;
        for (int i = 1; i <= 3; i++) {
            std::cout << "Wait 5s, then start reconnecting to kylin-ai-runtime."
                      << std::endl;
            sleep(5);
            std::cout << "start reconnect kylin-ai-runtime. rynum: " << i
                      << std::endl;
            initSessionRe = text_embedding_init_session(sessionRe);
            if (initSessionRe == 0) break;
            std::cout << "initSessionRe :" << initSessionRe << std::endl;
        }

        // 同步接口
        EmbeddingResult *resultRe = nullptr;
        text_embedding(sessionRe, "hello world!", &resultRe);
        float *vector_resultRe = embedding_result_get_vector_data(resultRe);
        int lenRe = embedding_result_get_vector_length(resultRe);
        std::cout << "vector_result lenRe: " << lenRe << std::endl;

        int error_codeRe = embedding_result_get_error_code(resultRe);
        std::cout << "error_codeRe: " << error_codeRe << std::endl;
        const char *error_messageRe =
            embedding_result_get_error_message(resultRe);
        std::cout << "error_messageRe :" << error_messageRe << std::endl;

        std::cout << "vector_result Re:" << std::endl;
        for (int i = 0; i < lenRe; i++) {
            std::cout << vector_resultRe[i] << ",";
        }
        std::cout << std::endl;
        embedding_result_destroy(&resultRe);
        text_embedding_destroy_session(&sessionRe);
    }
#endif

    std::cout << "释放结果资源" << std::endl;
    embedding_result_destroy(&result);
    embedding_model_info_destroy(info);
    text_embedding_destroy_session(&session);
}
void textEmbeddingAsync() {
    TextEmbeddingSession *session = text_embedding_create_session();
    if (text_embedding_init_session(session) != 0) {
        std::cout << "init session failed" << std::endl;
    }
    text_embedding_async(session, "热爱学习", callback, nullptr);
    auto *loop_ = g_main_loop_new(nullptr, false);
    std::thread ctrlThread([&session, loop_] {
        while (std::getchar() != '\n') {
        }
        g_main_loop_quit(loop_);
    });
    ctrlThread.detach();

    g_main_loop_run(loop_);
    text_embedding_destroy_session(&session);
}

std::vector<uint8_t> readFile(const std::string &filePath) {
    std::ifstream file(filePath, std::ios::binary | std::ios::ate);
    if (!file.is_open()) {
        throw std::runtime_error("Failed to open file");
    }

    std::streamsize size = file.tellg();
    file.seekg(0, std::ios::beg);

    std::vector<uint8_t> buffer(size);
    if (file.read(reinterpret_cast<char *>(buffer.data()), size)) {
        return buffer;
    } else {
        throw std::runtime_error("Failed to read file");
    }
}
const std::string base64_chars =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    "abcdefghijklmnopqrstuvwxyz"
    "0123456789+/";

std::string base64Encode(const std::vector<uint8_t> &buffer) {
    std::string encodedData;
    int i = 0;
    uint8_t char_array_3[3];
    uint8_t char_array_4[4];

    while (i < buffer.size()) {
        char_array_3[0] = buffer[i++];
        char_array_3[1] = (i < buffer.size()) ? buffer[i++] : 0;
        char_array_3[2] = (i < buffer.size()) ? buffer[i++] : 0;

        char_array_4[0] = (char_array_3[0] & 0xfc) >> 2;
        char_array_4[1] =
            ((char_array_3[0] & 0x03) << 4) + ((char_array_3[1] & 0xf0) >> 4);
        char_array_4[2] =
            ((char_array_3[1] & 0x0f) << 2) + ((char_array_3[2] & 0xc0) >> 6);
        char_array_4[3] = (char_array_3[2] & 0x3f);

        for (int j = 0; (j < 4); ++j) {
            encodedData += base64_chars[char_array_4[j]];
        }
    }

    while ((encodedData.size() % 4) != 0) {
        encodedData += '=';
    }

    return encodedData;
}
void imageEmbeddingSync() {
    ImageEmbeddingSession *session = image_embedding_create_session();

    if (image_embedding_init_session(session) != 0) {
        std::cout << "init session failed" << std::endl;
    }
    namespace fs = std::filesystem;
    fs::path dir = fs::path(__FILE__).parent_path();
    fs::path path = dir / "微信图片_20240709181353.jpg";

    std::string imagePath = path.string();
    std::vector<uint8_t> imageData = readFile(imagePath);
    std::string base64Data = base64Encode(imageData);
    unsigned int length = base64Data.length();
    const char *charPtr = base64Data.c_str();
    const unsigned char *ucharPtr =
        reinterpret_cast<const unsigned char *>(charPtr);

#if 0
    EmbeddingResult *result = nullptr;
    text_embedding_by_image_model(session, "do you love working?", &result);
#elif 1
    // 同步向量化图片
    EmbeddingResult *result = nullptr;
    image_embedding_by_image_file(session, path.string().c_str(),
                                  &result);  // 换成自己路径的图片
#else
    // 同步向量化base64图片

    EmbeddingResult *result = nullptr;
    image_embedding_by_base64_image_data(session, ucharPtr, length, &result);
#endif
    // 结果信息
    int error_code = embedding_result_get_error_code(result);
    std::cout << "error_code: " << error_code << std::endl;
    const char *error_message = embedding_result_get_error_message(result);
    std::cout << "error_message :" << error_message << std::endl;

    float *vector_result = embedding_result_get_vector_data(result);
    int len = embedding_result_get_vector_length(result);

    std::cout << "vector_result :" << std::endl;
    for (int i = 0; i < len; i++) {
        std::cout << vector_result[i] << " ";
    }
    std::cout << std::endl;

    char *info = nullptr;
    image_embedding_get_model_info(session, &info);
    std::cout << "print model info:" << info << std::endl;

#if 0
    // 测试重连
    GMainLoop *loop = g_main_loop_new(nullptr, false);
    g_timeout_add_seconds(5, quit_loop, loop);
    g_main_loop_run(loop);

    EmbeddingResult *result1 = nullptr;
    // text_embedding_by_image_model(session, "12345", &result1);
    image_embedding_by_image_file(session, path.string().c_str(),
                                  &result1);  // 换成自己路径的图片
    // image_embedding_by_base64_image_data(session, ucharPtr, length,
    // &result1);
    int error_code1 = embedding_result_get_error_code(result1);
    std::cout << "error_code1: " << error_code1 << std::endl;
    const char *error_message1 = embedding_result_get_error_message(result1);
    std::cout << "error_message1 :" << error_message1 << std::endl;

    if (error_code1 == COREAI_EMBEDDING_CONNECTION_ERROR) {
        // 错误码为10 说明连接runtime有问题
        embedding_result_destroy(&result);
        image_embedding_destroy_session(&session);
        //  由于runtime的问题，所以 重新创建session,重连两次，每次等待10s
        ImageEmbeddingSession *sessionRe = image_embedding_create_session();

        int initSessionRe;
        for (int i = 1; i <= 3; i++) {
            std::cout << "Wait 5s, then start reconnecting to kylin-ai-runtime."
                      << std::endl;
            sleep(5);
            std::cout << "start reconnect kylin-ai-runtime. rynum: " << i
                      << std::endl;
            initSessionRe = image_embedding_init_session(sessionRe);
            if (initSessionRe == 0) break;
            std::cout << "initSessionRe :" << initSessionRe << std::endl;
        }
        if (initSessionRe == COREAI_EMBEDDING_CONNECTION_ERROR) return;
        // 同步接口
        embedding_result_destroy(&result1);

        text_embedding_by_image_model(sessionRe, "hello world!", &result1);
        // image_embedding_by_image_file(sessionRe, path.string().c_str(),
        //                               &result1);  // 换成自己路径的图片
        // image_embedding_by_base64_image_data(sessionRe, ucharPtr, length,
        //                                      &result1);

        float *vector_resultRe = embedding_result_get_vector_data(result1);
        int lenRe = embedding_result_get_vector_length(result1);
        std::cout << "vector_result lenRe: " << lenRe << std::endl;

        int error_codeRe = embedding_result_get_error_code(result1);
        std::cout << "error_codeRe: " << error_codeRe << std::endl;
        const char *error_messageRe =
            embedding_result_get_error_message(result1);
        std::cout << "error_messageRe :" << error_messageRe << std::endl;

        std::cout << "vector_result Re:" << std::endl;
        for (int i = 0; i < lenRe; i++) {
            std::cout << vector_resultRe[i] << ",";
        }
        std::cout << std::endl;
        embedding_result_destroy(&result1);
        image_embedding_destroy_session(&sessionRe);
    }
#endif

    std::cout << "释放结果资源" << std::endl;
    embedding_result_destroy(&result);
    embedding_model_info_destroy(info);
    image_embedding_destroy_session(&session);
}

void imageEmbeddingAsync() {
    ImageEmbeddingSession *session = image_embedding_create_session();

    if (image_embedding_init_session(session) != 0) {
        std::cout << "init session failed" << std::endl;
    }

    namespace fs = std::filesystem;
    fs::path dir = fs::path(__FILE__).parent_path();
    fs::path path = dir / "微信图片_20240709181353.jpg";

    std::string imagePath = path.string();
    std::vector<uint8_t> imageData = readFile(imagePath);
    std::string base64Data = base64Encode(imageData);
    unsigned int length = base64Data.length();
    const char *charPtr = base64Data.c_str();
    const unsigned char *ucharPtr =
        reinterpret_cast<const unsigned char *>(charPtr);

#if 0
    text_embedding_by_image_model_async(session,"do you love working?",callback,nullptr);
#elif 0
    image_embedding_from_by_file_async(session, imagePath.c_str(), callback,
                                       nullptr);
#else
    image_embedding_by_base64_image_data_async(session, ucharPtr, length,
                                               callback, nullptr);
#endif

    auto *loop_ = g_main_loop_new(nullptr, false);
    std::thread ctrlThread([&session, loop_] {
        while (std::getchar() != '\n') {
        }
        g_main_loop_quit(loop_);
    });
    ctrlThread.detach();

    g_main_loop_run(loop_);
    image_embedding_destroy_session(&session);
}

void threadTask(TextEmbeddingSession *session, const std::string &input,
                int threadId) {
    std::cout << "threadId id: " << threadId << std::endl;
    EmbeddingResult *result = nullptr;
    bool sucess = text_embedding(session, input.c_str(), &result);

    // 结果信息
    int error_code = embedding_result_get_error_code(result);
    std::cout << "error_code: " << error_code << std::endl;
    const char *error_message = embedding_result_get_error_message(result);
    std::cout << "error_message :" << error_message << std::endl;

    float *vector_result = embedding_result_get_vector_data(result);
    int len = embedding_result_get_vector_length(result);

    std::cout << "vector_result :" << std::endl;
    for (int i = 0; i < len; i++) {
        std::cout << vector_result[i] << ",";
    }
    std::cout << std::endl;
    std::cout << "=======================================" << std::endl;
}
void test01() {
    // 测试单例连接的问题
    TextEmbeddingSession *session = text_embedding_create_session();
    int initSession = text_embedding_init_session(session);
    if (initSession != 0) {
        std::cout << "init session failed" << std::endl;
    }
    std::cout << "initSession :" << initSession << std::endl;

    std::thread thread1(threadTask, session, "input1", 1);
    sleep(5);
    std::thread thread2(threadTask, session, "input2", 2);

    thread1.join();
    thread2.join();
}
int main(int argc, char *argv[]) {
    textEmbeddingSync();
    // test01();

    // textEmbeddingAsync();
    // imageEmbeddingSync();
    // imageEmbeddingAsync();
    fflush(stdout);

    return 0;
}